#include "ros/ros.h"
#include "std_msgs/MultiArrayLayout.h"
#include "std_msgs/MultiArrayDimension.h"
#include "std_msgs/Int8.h"
#include "std_msgs/Int16.h"
#include "ar_track_alvar/AlvarMarkers.h"
#include "ar_track_alvar/AlvarMarker.h"
#include <queue> 
#include <tf/LinearMath/Quaternion.h> // Needed to convert rotation ...
#include <tf/LinearMath/Matrix3x3.h>  // ... quaternion into Euler angles
#include <tf/transform_listener.h>
#include <tf/tf.h>

using namespace std;

//Constants
#define PI 3.14159265
#define CONTROL_HZ 20
#define NUM_MARKERS 1
#define MAX_VALID_MARKERS 5
#define MAX_INVALID_MARKERS 5
#define QUEUE_SIZE 10
#define EPSILON 0.1
#define X_MIN_DISTANCE 0.5//0.686
#define RAMP_LOCATION_X_MIN 0.432
#define RAMP_LOCATION_X_MAX 0.5//0.686
#define RAMP_LOCATION_Y_MIN -0.1//-0.165
#define RAMP_LOCATION_Y_MAX 0.1//0.165
#define MAX_VELOCITY 150
#define MIN_VELOCITY 0
#define VELOCITY_NOMINAL 100
#define CORRECTION_TERM_LEFT 55
#define CORRECTION_TERM_RIGHT 55
#define CLOCKWISE 1
#define ANTICLOCKWISE -1
#define MOVE_STRAIGHT_PERSIST_TIME 20
#define TURN_IN_PLACE_PERSIST_TIME 0
#define OBJECT_SEARCH_PERSIST_TIME 5
#define FINAL_ALIGN_PERSIST_TIME 5
#define PAUSE_PERSIST_TIME 20
#define TURN_RATE 0.8 //0.278
#define STRAIGHT_RATE 0.5
#define INITIALIZATION_TIME 25
#define DEG2RAD(X) (X*PI/180)
#define RAD2DEG(X) (X*180/PI)
#define EPS_ANGLE 5
#define LEFT_SIDE 1
#define RIGHT_SIDE -1
#define STRAIGHT_SIDE 0
#define MAX_OBJECT_SEARCH_DURATION 40
#define HAIL_MARY_DURATION 80
#define HAIL_MARY_VELOCITY 255


//Datastructure
class position_marker {
public:
  position_marker()	{
    x_pos = 0;
    y_pos = 0;
    yaw = 0;
    isValid = false;
  }
  double x_pos;
  double y_pos;
  double yaw;
  bool isValid;
};

//Publishers and subscribers
ros::Publisher velocity_pub;
ros::Subscriber position_sub;
tf::TransformListener *listener;
tf::StampedTransform *curr_transform;

//Variables
int curr_valid_markers = 0;
int curr_invalid_markers = 0;
bool is_ramp_reached    = false;
int velocity_right = 0;
int velocity_left = 0;
bool turn_in_place = false;
bool move_reverse = false;
bool move_forward = false;
int turn_in_place_dir = CLOCKWISE;
int action_persist_counter = 0;
bool action_persist = false;
bool action_persist_override = false;
int persist_time = 0;
int next_persist_time = 0;
int initialization_counter = 0;
int robot_side = 0;
int objectSearchCounter = 0;
bool objectLost = false;
bool finalAlign = false;
int last_robot_side = 0;
bool hailMaryRun = false;
int hail_mary_counter = 0;
bool goalAlignRun = false;

deque <position_marker> pos_queue;
std_msgs::Int16 vel;

//Functions
void controlCallback(const ros::TimerEvent& t);
void positionCallback(const ar_track_alvar::AlvarMarkers::ConstPtr& pos_markers);
void getPosition( position_marker& curr_pos, bool& posAvailable );
void simpleCompassController();
void sendVelocityCmd( int velocity_right, int velocity_left );
void stopWegRobot();
void turnInPlaceController();
void moveReverseController();
void moveForwardController();
void angleCalc(position_marker& curr_pos, bool& posAvailable, bool& temp_turn_in_place, double& diff_angle);
void goalAlign( position_marker& curr_pos, double& diff_angle);
int calcRobotSide (position_marker& curr_pos);
bool updateFlightStatus( double& yaw, double& pitch, double& roll );
void climbRamp();

int main(int argc, char **argv)
{
  ros::init(argc, argv, "weg_controller");
  ros::NodeHandle n;
  
  listener = new tf::TransformListener();
  curr_transform = new tf::StampedTransform();

  velocity_pub = n.advertise<std_msgs::Int16>("weg_vel", 10);
  ros::Subscriber position_sub = n.subscribe("ar_pose_marker", 10, positionCallback);
  ros::Timer timer = n.createTimer(ros::Duration(1.0/CONTROL_HZ), controlCallback);

  //Ros loop
  ros::Rate loop_rate(10);
  while (ros::ok()) {
    ros::spinOnce();
    //loop_rate.sleep();
  }

  return 0;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Control callback function

void controlCallback(const ros::TimerEvent& t)
{
  bool posAvailable = false;
  position_marker curr_pos;
  bool prev_move_forward, prev_move_reverse;
  bool temp_turn_in_place = false;
  double diff_angle = 0;
  
  if(hailMaryRun)	{
    if(++hail_mary_counter < HAIL_MARY_DURATION)	{
      climbRamp();
    } else {
      cout<<"Woohoo! climbed ramp"<<endl;
      stopWegRobot();
      exit(1);
    }
  }
  
  if(goalAlignRun)	{
      angleCalc(curr_pos, posAvailable, temp_turn_in_place, diff_angle);
      cout<<"Diff_angle "<<diff_angle<<endl;
      cout<<"Robot_side "<<robot_side<<endl;
      if(temp_turn_in_place)	{
	cout<<"Correcting alignment to goal"<<endl;
	goalAlign(curr_pos, diff_angle);
	
      } else {
	cout<<"Goal alignment complete!"<<endl;
	stopWegRobot();
	goalAlignRun = false;
	hailMaryRun = true;
	//exit(1);
      }
  }
  
  if( !is_ramp_reached )	{
    
    //Initialization time
    if(++initialization_counter < INITIALIZATION_TIME)		{
      //cout<<"Initializing ..."<<endl;
      return;
    }
        
    //Get current robot position
    //cout<<"Acquiring position ..."<<endl;
    getPosition( curr_pos, posAvailable );
    cout<<"Aquired Pos Index,X,Y,YAW:  "<<posAvailable<<"\t"
    <<curr_pos.x_pos<<"\t"<<curr_pos.y_pos<<"\t"<<curr_pos.yaw<<endl;

    if(posAvailable)	{
      last_robot_side = calcRobotSide( curr_pos );
      if(objectLost) 	stopWegRobot();
      objectLost = false;
      objectSearchCounter = 0;
    }
    //Check if robot is in target position
    if( posAvailable &&
	curr_pos.x_pos > RAMP_LOCATION_X_MIN - EPSILON &&
        curr_pos.x_pos < RAMP_LOCATION_X_MAX + EPSILON &&
        curr_pos.y_pos > RAMP_LOCATION_Y_MIN &&
        curr_pos.y_pos < RAMP_LOCATION_Y_MAX )	{
      is_ramp_reached = true;
      cout<<"Target position reached"<<endl;  
    
      goalAlignRun = true;
      stopWegRobot();
     
    //Check if action persist override and turn in place
    } else if (action_persist_override && turn_in_place)	{
	if(++action_persist_counter > persist_time)	{
	  cerr<<"Turn in place completed"<<endl;
	  prev_move_forward = move_forward;
	  prev_move_reverse = move_reverse;
	  stopWegRobot();
	  action_persist = true;
	  move_forward = prev_move_forward;
	  move_reverse = prev_move_reverse;
	  
	  if(move_forward || move_reverse)
	    action_persist_override = true;
	  
	  persist_time = next_persist_time;
	  turn_in_place_dir = ( turn_in_place_dir == CLOCKWISE) ? ANTICLOCKWISE : CLOCKWISE;
	  
	} else {
	  cout<<"Turning in place persist ("<<turn_in_place_dir<<")"<<endl;
	  turnInPlaceController();
	}

    //Check if action persist override and move reverse
    } else if (action_persist_override && move_reverse)	{
	if(++action_persist_counter > persist_time)	{
	  cerr<<"Move reverse completed"<<endl;
	  stopWegRobot();
	  
	} else {
	  cout<<"Move reverse persist"<<endl;
	  moveReverseController();
	}

    //Check if action persist override and move forward
    } else if (action_persist_override && move_forward)	{
	if(++action_persist_counter > persist_time)	{
	  cerr<<"Move forward completed"<<endl;
	  stopWegRobot();
	  
	} else {
	  cout<<"Move forward persist"<<endl;
	  moveForwardController();
	}
	
    //Check if action persist is active and tag found and no override
    } else if ( posAvailable && action_persist && !action_persist_override)	{
      cout<<"Tag found during action persist"<<endl;
      stopWegRobot();
      //exit(1);
            
    //Check if action persist is active and tag not found
    } else if ( !posAvailable && action_persist )	{
      
      //Turning in place
      if(turn_in_place)		{
	if(++action_persist_counter > persist_time)	{
	  stopWegRobot();
	  
	  if(objectSearchCounter < MAX_OBJECT_SEARCH_DURATION && objectLost)	{
	        turn_in_place = true;
		turn_in_place_dir = CLOCKWISE;
		action_persist_counter = 1;
		action_persist = true;
		objectSearchCounter++;
		persist_time = OBJECT_SEARCH_PERSIST_TIME;
		turnInPlaceController();
	  } else {
		cerr<<"Target not found: Turn in place time out"<<endl;
		return;
	  }
	} else {
	  cout<<"Turning in place persist"<<endl;
	  turnInPlaceController();
	}
	
      //Reverse
      } else if (move_reverse)		{
	if(++action_persist_counter > persist_time)	{
	  cout<<"Move reverse finished"<<endl;
	  stopWegRobot();
	  
	} else {
	  cout<<"Move reverse persist"<<endl;
	  moveReverseController();
	}

      //Forward
      } else if (move_forward)		{
	if(++action_persist_counter > persist_time)	{
	  cout<<"Move forward finished"<<endl;
	  stopWegRobot();
	  
	} else {
	  cout<<"Move forward persist"<<endl;
	  moveForwardController();
	}
	
      //Unknown action persist
      } else {
	cerr<<"Unknown persist error"<<endl;
	stopWegRobot();
	exit(1);
      }
    
    //Position available and goal not reached, so compass controller
    } else if (posAvailable)	{ 
      cout<<"Initiating simple compass controller ..."<<endl;
      simpleCompassController();
      
    //Unknown control path  
    } else {
      cout<<"Target lost, looking in circles"<<endl;
      simpleCompassController();
      //cerr<<"Unknown control path"<<endl;
      //stopWegRobot();
      //exit(1);
    }

  } 
  
  //Sending control velocity
  cout<<"Sending velocity (right, left):\t"<<velocity_right<<"\t"<<velocity_left<<endl;
  sendVelocityCmd( velocity_right, velocity_left );
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Position callback function
 
void positionCallback(const ar_track_alvar::AlvarMarkers::ConstPtr& pos_markers) {
  position_marker temp_marker;
  ar_track_alvar::AlvarMarker temp_alvar_marker;
  double roll = 0, pitch = 0, yaw = 0, x_pos = 0, y_pos = 0;
  
  if (pos_markers->markers.size() == NUM_MARKERS)	{
    temp_alvar_marker = pos_markers->markers[0];
    
    temp_marker.isValid = true;
    
    if(updateFlightStatus( x_pos, y_pos, yaw ))	{
      tf::Quaternion q(temp_alvar_marker.pose.pose.orientation.x, 
		      temp_alvar_marker.pose.pose.orientation.y,
		      temp_alvar_marker.pose.pose.orientation.z,
		      temp_alvar_marker.pose.pose.orientation.w);
      tf::Matrix3x3(q).getRPY(roll, pitch, yaw);

      temp_marker.x_pos = temp_alvar_marker.pose.pose.position.x;
      temp_marker.y_pos = temp_alvar_marker.pose.pose.position.y;
      temp_marker.yaw = yaw;
    } else {
      temp_marker.x_pos = x_pos;
      temp_marker.y_pos = y_pos;
      temp_marker.yaw = yaw;
    }
    
    //std::cout << "X: " << temp_marker.x_pos << ", Y: " << temp_marker.y_pos 
    //<< ", Roll: " << roll <<", Pitch: " << pitch << ", Yaw: " << yaw << endl;
  }
  
  //Add points to queue
  if(pos_queue.size() > QUEUE_SIZE)	{
    pos_queue.pop_front();
    pos_queue.push_back(temp_marker);
  } else {
    pos_queue.push_back(temp_marker);
  }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Get position function

void getPosition( position_marker& curr_pos, bool& posAvailable ) {
  
  int numValidPts = 0;
  int numInvalidPts = 0;
  double curr_x_pos = 0;
  double curr_y_pos = 0;
  double curr_yaw = 0;
  
  deque <position_marker> temp_queue = pos_queue;
  
  if(temp_queue.size() > QUEUE_SIZE)	{
    for (deque<position_marker>::iterator i = temp_queue.begin(); i != temp_queue.end() ; ) {     
      if(numValidPts+numInvalidPts == QUEUE_SIZE)	{
	break;
      }
      if (i->isValid)	{
	numValidPts ++;
	curr_x_pos += i->x_pos;
	curr_y_pos += i->y_pos;
	curr_yaw += (RAD2DEG(i->yaw));
	//cout<<"valid point available .. \t"<<i->x_pos<<"\t"<<i->y_pos<<"\t"<<i->yaw<<"\t"<<RAD2DEG(i->yaw)<<endl;
      } else {
	numInvalidPts ++;
	//cout<<"invalid point available .."<<endl;
      }
    }
  } else {
    cerr<<"Not enough messages in queue"<<endl;
    posAvailable = false;
  }
    
  if(numValidPts >= MAX_VALID_MARKERS)	{
    curr_pos.x_pos = curr_x_pos / numValidPts;
    curr_pos.y_pos = curr_y_pos / numValidPts;
    curr_pos.yaw = curr_yaw / numValidPts;
    posAvailable = true;
  } else {
    posAvailable = false;
  }
  
  //cout<<"Moving Avg  X,Y,Yaw:\t"<<curr_pos.x_pos<<"\t"<<curr_pos.y_pos<<"\t"<<curr_pos.yaw<<endl;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Simple compass type controller

void simpleCompassController()	{
  bool posAvailable = false;
  bool temp_turn_in_place = false;
  double diff_angle = 0;
  
  position_marker curr_pos;

  //Get current robot position
  getPosition( curr_pos, posAvailable );
    
  if(posAvailable)	{
    last_robot_side = calcRobotSide( curr_pos );
    objectLost = false;
    objectSearchCounter = 0;
  }
  
  //Computing angles and turn in place directions
  angleCalc(curr_pos, posAvailable, temp_turn_in_place, diff_angle);  
  
  if (posAvailable && curr_pos.x_pos < X_MIN_DISTANCE)	{
    if (temp_turn_in_place)	{
      //turn_in_place_dir = CLOCKWISE;
      turn_in_place = true;
      move_forward = true;
      persist_time = TURN_RATE * 3 * fabs(diff_angle);
      next_persist_time = MOVE_STRAIGHT_PERSIST_TIME;
      action_persist_override = true;
      cout<<"Turning in "<<turn_in_place_dir<<" with persist time "<<persist_time
	  <<" Move forward time "<<next_persist_time
	  <<" Diff angle "<<diff_angle<<endl;
      return;
    } else {
      cout<<"No turn in place and X distance is upclose to target"<<endl;
      return;
    }
      /*else {
      //turn_in_place_dir = ANTICLOCKWISE;
      turn_in_place = true;
      move_forward = true;
      persist_time = TURN_RATE * 3 * fabs(diff_angle); 
      next_persist_time = MOVE_STRAIGHT_PERSIST_TIME;
      action_persist_override = true;
      cout<<"Turning in "<<turn_in_place_dir<<" with persist time "<<persist_time
	  <<" Diff angle "<<diff_angle<<endl;
    }*/

  } else if (posAvailable && curr_pos.x_pos >= X_MIN_DISTANCE)	{
    if (temp_turn_in_place)	{
      //turn_in_place_dir = CLOCKWISE;
      turn_in_place = true;
      move_forward = true;
      persist_time = TURN_RATE * fabs(diff_angle);
      next_persist_time = MOVE_STRAIGHT_PERSIST_TIME;//STRAIGHT_RATE * fabs(curr_pos.y_pos);
      action_persist_override = true;
      cout<<"Turning in "<<turn_in_place_dir<<" with persist time "<<persist_time
      <<" Move forward time "<<next_persist_time
	  <<" Diff angle "<<diff_angle<<endl;
      return;	  
    } else {
      cout<<"No turn in place and X distance is away from target"<<endl;
    }
    /*else {
      //turn_in_place_dir = ANTICLOCKWISE;
      turn_in_place = true;
      move_forward = true;
      persist_time = TURN_RATE * fabs(diff_angle); 
      next_persist_time = STRAIGHT_RATE * fabs(curr_pos.y_pos);
      action_persist_override = true;
      cout<<"Turning in "<<turn_in_place_dir<<" with persist time "<<persist_time
	  <<" Diff angle "<<diff_angle<<endl;
    }*/   
  }
    
  if (posAvailable && (diff_angle < 2*EPS_ANGLE && diff_angle > -2*EPS_ANGLE) && robot_side == LEFT_SIDE)	{ 
        velocity_left = VELOCITY_NOMINAL + CORRECTION_TERM_LEFT;
	velocity_right = VELOCITY_NOMINAL;
	cout<<"Homing from left"<<endl;
  } else if (posAvailable && (diff_angle < 2*EPS_ANGLE && diff_angle > -2*EPS_ANGLE) && robot_side == RIGHT_SIDE)	{
        velocity_left = VELOCITY_NOMINAL;
	velocity_right = VELOCITY_NOMINAL + CORRECTION_TERM_RIGHT;  
	cout<<"Homing from right"<<endl;
  } else if (posAvailable)	{
    velocity_left = VELOCITY_NOMINAL;
    velocity_right = VELOCITY_NOMINAL;
    cout<<"Homing straight"<<endl;
  } else {
    turn_in_place = true;
    if (last_robot_side == LEFT_SIDE)
      turn_in_place_dir = ANTICLOCKWISE;
    else
      turn_in_place_dir = CLOCKWISE;
    action_persist_counter = 1;
    action_persist = true;
    if(!objectLost)
      objectSearchCounter = 0;
    else
      objectLost = true;
    persist_time = OBJECT_SEARCH_PERSIST_TIME;
    turnInPlaceController();
    cout<<"Target lost, going aroung in circles"<<endl;
  }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Send velocity command

void sendVelocityCmd( int vel_right, int vel_left )	{
  
  int velocity_right_sgn = 0, velocity_left_sgn = 0;
  int velocity_right_mag = 0, velocity_left_mag = 0;
  
  velocity_right_sgn = (vel_right>0) ? 1:-1;
  velocity_left_sgn = (vel_left>0) ? 1:-1;
  velocity_right_mag = fabs(vel_right);
  velocity_left_mag = fabs(vel_left);
  
  if (velocity_right_mag > MAX_VELOCITY)	velocity_right_mag = MAX_VELOCITY;
  if (velocity_left_mag > MAX_VELOCITY)	velocity_left_mag = MAX_VELOCITY;
  if (velocity_right_mag < MIN_VELOCITY)	velocity_right_mag = MIN_VELOCITY;
  if (velocity_left_mag < MIN_VELOCITY)	velocity_left_mag = MIN_VELOCITY;
  
  velocity_right = velocity_right_mag * velocity_right_sgn;
  velocity_left = velocity_left_mag * velocity_left_sgn;
  
  //Right side velocity
  if(velocity_right>0)
    vel.data = 2000+fabs(velocity_right);
  else
    vel.data = 4000+fabs(velocity_right);
  
  //Sending right velocity
  velocity_pub.publish(vel);
  //cout<<"Sending left velocity = "<<vel.data<<endl;
  
  //Left side velocity
  if(velocity_left>0)
    vel.data = 1000+fabs(velocity_left);
  else
    vel.data = 3000+fabs(velocity_left);
  
  //Sending left velocity
  velocity_pub.publish(vel);
  //cout<<"Sending right velocity = "<<vel.data<<endl;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Stop gracefully

void stopWegRobot()	{
    action_persist = false;
    action_persist_override = false;
    turn_in_place = false;
    move_reverse = false;
    move_forward = false;
    action_persist_counter = 0;
    sendVelocityCmd( 0, 0 );
} 

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Turn in place controller

void turnInPlaceController()	{
  if(turn_in_place_dir == CLOCKWISE)	{
    velocity_left = MAX_VELOCITY;
    velocity_right = -MAX_VELOCITY;
  } else {
    velocity_left = -MAX_VELOCITY;
    velocity_right = MAX_VELOCITY;
  } 
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Move reverse controller

void moveReverseController()	{
  velocity_right = -VELOCITY_NOMINAL;
  velocity_left = -VELOCITY_NOMINAL;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Move forward controller

void moveForwardController()	{
  velocity_right = VELOCITY_NOMINAL;
  velocity_left = VELOCITY_NOMINAL;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Relative angles computation

void angleCalc(position_marker& curr_pos, bool& posAvailable, bool& temp_turn_in_place, double& diff_angle)	{

  if(posAvailable)	{
    double conn_line_angle = RAD2DEG( atan( fabs(curr_pos.x_pos - (RAMP_LOCATION_X_MAX+RAMP_LOCATION_X_MIN)/2) /
					fabs(curr_pos.y_pos) ) );
    
    double yaw_angle = (curr_pos.yaw > -90) ? -curr_pos.yaw:180+curr_pos.yaw;
    diff_angle = yaw_angle-conn_line_angle;
    
    cout<<"Angle calc parameters"<<endl;
    cout<<"X line = "<<fabs(curr_pos.x_pos - (RAMP_LOCATION_X_MAX+RAMP_LOCATION_X_MIN)/2)<<endl;
    cout<<"Y line = "<<fabs(curr_pos.y_pos) <<endl;
    cout<<"Connection line angle = "<<conn_line_angle<<endl;
    cout<<"Yaw angle = "<<yaw_angle<<endl;
    cout<<"Diff angle = "<<diff_angle<<endl;

    robot_side = calcRobotSide (curr_pos);
    
    cout<<"Robot Side  "<<robot_side<<endl;
    /*if(curr_pos.yaw < 0 + EPS_ANGLE && curr_pos.yaw > -(90-2*EPS_ANGLE))	{
      robot_side = LEFT_SIDE;
    } else if(curr_pos.yaw < -(90 + 2*EPS_ANGLE) && curr_pos.yaw > -(180 - EPS_ANGLE))	{
      robot_side = RIGHT_SIDE;
    } else if(curr_pos.yaw <= -(90 - 2*EPS_ANGLE) && curr_pos.yaw >= -(90 + 2*EPS_ANGLE))	{
      robot_side = STRAIGHT_SIDE;
    } else {
      cerr<<"Unknown Angle in compass"<<endl;
    }*/
    
    if(diff_angle < 2*EPS_ANGLE && diff_angle > -2*EPS_ANGLE)	{
      return;
    }
    
    if(diff_angle > 0 && robot_side == LEFT_SIDE)	{
      turn_in_place_dir = CLOCKWISE;
      temp_turn_in_place = true;
    } else if(diff_angle <= 0 && robot_side == LEFT_SIDE) {
      turn_in_place_dir = ANTICLOCKWISE;
      temp_turn_in_place = true;
    } else if(diff_angle > 0 && robot_side == RIGHT_SIDE) {
      turn_in_place_dir = ANTICLOCKWISE;
      temp_turn_in_place = true;
    } else if(diff_angle <= 0 && robot_side == RIGHT_SIDE) {
      turn_in_place_dir = CLOCKWISE;
      temp_turn_in_place = true;
    }
  }  
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Final goal alignment

void goalAlign( position_marker& curr_pos, double& diff_angle)	{
  //turn_in_place_dir = CLOCKWISE;
  //turn_in_place = true;
  //move_forward = false;
  //persist_time = FINAL_ALIGN_PERSIST_TIME;
  //next_persist_time = 0;
  //action_persist_override = false;
  cout<<"Turning in "<<turn_in_place_dir<<" with persist time "<<persist_time
      <<" Diff angle "<<diff_angle<<endl;
  turnInPlaceController();
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Calculate which side of target is the robot

int calcRobotSide (position_marker& curr_pos)	{
  int curr_robot_side = 0;
    if(curr_pos.yaw < 0 + EPS_ANGLE && curr_pos.yaw > -(90-2*EPS_ANGLE))	{
      curr_robot_side = LEFT_SIDE;
    } else if(curr_pos.yaw < -(90 + 2*EPS_ANGLE) && curr_pos.yaw > -(180 - EPS_ANGLE))	{
      curr_robot_side = RIGHT_SIDE;
    } else if(curr_pos.yaw <= -(90 - 2*EPS_ANGLE) && curr_pos.yaw >= -(90 + 2*EPS_ANGLE))	{
      curr_robot_side = STRAIGHT_SIDE;
    } else {
      cerr<<"Unknown Angle in compass"<<endl;
    }
    return(curr_robot_side);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Calculate which side of target is the robot

bool updateFlightStatus( double& x_pos, double& y_pos, double& yaw )
{
  //cout<<"Trying to get transform"<<endl;
  double curr_roll = 0,curr_pitch = 0,curr_yaw = 0, curr_x_pos = 0, curr_y_pos = 0, curr_z_pos = 0;

  try{
    listener->lookupTransform("camera_link", "ar_marker_0", ros::Time(0), *curr_transform);
  }
  catch (tf::TransformException ex){
    ROS_ERROR("%s",ex.what());
    cout<<"No transform found"<<endl;
    return(true);
  }

  //get angles in ZYX
  tf::Matrix3x3(curr_transform->getRotation()).getEulerYPR(curr_yaw,curr_pitch,curr_roll);

  //getting positions x,y,z
  curr_x_pos = curr_transform->getOrigin().x();
  curr_y_pos = curr_transform->getOrigin().y();
  curr_z_pos = curr_transform->getOrigin().z();

  x_pos = curr_x_pos;
  y_pos = curr_y_pos;
  yaw = curr_yaw;
  
  //cout<<"Transform found"<<endl;
  return(false);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//Hail mary go straight

void climbRamp()	{
  cout<<"Hail Mary Run"<<endl;
  sendVelocityCmd( HAIL_MARY_VELOCITY, HAIL_MARY_VELOCITY);
}